/*
 * Copyright (C) 2019 CERN (www.sevensols.com)
 * Author: Miguel Jimenez Lopez <miguel.jimenez@sevensols.com>
 *
 * Released according to the GNU GPL, version 2 or any later version.
 *
 * This work is part of the White Rabbit project, a research effort led
 * by CERN, the European Institute for Nuclear Research.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/firmware.h>
#include <linux/delay.h>
#include <linux/fmc.h>
#include <linux/fmc-sdb.h>

#include "fmc-dio.h"

// IRQ domain for SPEC board
static const char *spec_irqdomain_name = "htvic-spec.0";

/**
 * FMC DIO Interrupt Service Routing (ISR)
 *
 * @param irq IRQ number
 * @param dev_id Device identifier for the ISR
 *
 * @return IRQ_HANDLED if interrupt was handled properly and IRQ_NONE otherwise
 */
irqreturn_t fmc_dio_interrupt(int irq, void *dev_id)
{
	return fmc_dio_int_interrupt(dev_id);
}

// Macros for the memory-mapped hardware resources
#define FMC_DIO_RESOURCE_DIO 0
#define FMC_DIO_RESOURCE_GPIO 1
#define FMC_DIO_RESOURCE_PPSG 2
/**
 * This function creates the hardware resources for the FMC DIO.
 *
 * @param dio FMC DIO device
 *
 * @return 0 if success and a negative error code otherwise
 */
static int fmc_dio_resource_map(struct fmc_dio *dio)
{
	int i,err;
	struct platform_device *dev = dio->pdev;
	struct fmc_device *fmc = dio->fmc;
	struct resource *r;
	int irq;
	void __iomem *mem;
	struct irq_domain *irqdomain;

	// Map the resources
	for(i = 0 ; i < dev->num_resources ; i++) {
		r = platform_get_resource(dev, IORESOURCE_MEM, i);
		if(!r)
			continue;

		mem =  fmc->fpga_base + r->start;

		switch(i)
		{
		case FMC_DIO_RESOURCE_DIO:
			dio->dio = mem;
			break;
		case FMC_DIO_RESOURCE_GPIO:
			dio->gpio = mem;
			break;
		case FMC_DIO_RESOURCE_PPSG:
			dio->ppsg = mem;
			break;
		default:
			/* Ugh? What is this? */
			break;
		};
	}

	// If some hardware resource is not detected, fail
	if(!dio->dio || !dio->gpio || !dio->ppsg)
		return -EINVAL;

	// Search the IRQ domain
	irqdomain = irq_find_host((struct device_node *)dio->irqdomain_name);
	if (!irqdomain)
		return -EINVAL;

	// Map the IRQ
	irq = platform_get_irq(dev,0);
	if(irq < 0)
		return -EINVAL;

	// Register the IRQ handler
	dio->irq = irq_find_mapping(irqdomain, irq);
	err = request_irq(irq, fmc_dio_interrupt,
			  IRQF_TRIGGER_LOW, dio->name, dio);
	if(err)
		return err;

	return 0;
}

/**
 * This function releases the hardware resources for the FMC DIO.
 *
 * @param dio FMC DIO device
 *
 * @return 0 if success and a negative error code otherwise
 */
static int fmc_dio_resource_release(struct fmc_dio *dio)
{
	// Release IRQ for FMC DIO
	free_irq(dio->irq, dio);

	return 0;
}

/**
 * FMC DIO probe function
 *
 * This function initializes the FMC DIO data structures.
 *
 * @param fmc FMC DIO device
 *
 * @return 0 if success and a negative error code otherwise
 */
int fmc_dio_probe(struct fmc_device *fmc)
{
	struct fmc_dio *dio;

	/* Driver data */
	dio = devm_kzalloc(&fmc->dev, sizeof(*dio), GFP_KERNEL);
	if (!dio)
		return -ENOMEM;
	fmc_set_drvdata(fmc, dio);

	/* Set up the FMC DIO name for the Linux Miscdevice */
	snprintf(dio->name,FMC_DIO_NAME_MAX,FMC_DIO_NAME_PATTERN,fmc->device_id);
	dio->fmc = fmc;

	return 0;
}

/**
 * FMC DIO remove function
 *
 * This function releases the FMC DIO data structures.
 *
 * @param fmc FMC DIO device
 *
 * @return 0 if success and a negative error code otherwise
 */
int fmc_dio_remove(struct fmc_device *fmc)
{
	struct fmc_dio *dio = fmc_get_drvdata(fmc);

	/* Free FMC DIO internal structure */
	kfree(dio);

	return 0;
}

// FMC DIO FRU identifier
static struct fmc_fru_id fd_fru_id[] = {
	{
		.product_name = "FmcDio5cha",
	},
};

// FMC DIO driver structure (FMC)
static struct fmc_driver fmc_dio_drv = {
	.version = FMC_VERSION,
	.driver.name = KBUILD_MODNAME,
	.probe = fmc_dio_probe,
	.remove = fmc_dio_remove,
	.id_table = {
		.fru_id = fd_fru_id,
		.fru_id_nr = ARRAY_SIZE(fd_fru_id),
	},
};

/**
 * FMC DIO probe function for the platform_device
 *
 * @param dev FMC DIO platform device
 *
 * @return 0 if success and a negative error code otherwise
 */
static int fmc_dio_pl_probe(struct platform_device *dev)
{
	/* NOTE: fmc_device is created by the same driver than platform_device
	 * and it had filled the private data structure pointer of the
	 * platform_device before registering it.
	 */
	struct fmc_device *fmc = platform_get_drvdata(dev);
	struct fmc_dio *dio = fmc_get_drvdata(fmc);
	int ret = 0;

	// Specific board info
	dio->board = dev->id_entry->driver_data;

	// Assign IRQ domain depending on the specific board
	if(dio->board == FMC_DIO_BOARD_SPEC)
		dio->irqdomain_name = spec_irqdomain_name;
	// FIXME: Implement for other boards

	/* Map the resources for the FMC DIO */
	ret = fmc_dio_resource_map(dio);
	if(ret)
		return ret;

	/* Create a Linux Miscdevice for the FMC DIO */
	ret = fmc_dio_mdev_create(dio);
	if(ret)
		return ret;

	/* Create Linux GPIO for the FMC DIO */
	ret = fmc_dio_gpio_init(dio);

	return ret;
}

/**
 * FMC DIO remove function for the platform_device
 *
 * @param dev FMC DIO platform device
 *
 * @return 0 if success and a negative error code otherwise
 */
static int fmc_dio_pl_remove(struct platform_device *dev)
{
	struct fmc_device *fmc = platform_get_drvdata(dev);
	struct fmc_dio *dio = fmc_get_drvdata(fmc);
	int ret = 0;

	/* Release Linux GPIO for the FMC DIO */
	fmc_dio_gpio_exit(dio);

	/* Release the resources for the FMC DIO */
	ret = fmc_dio_resource_release(dio);
	if(ret)
		return ret;

	/* Destroy the Linux Miscdevice for the FMC DIO */
	fmc_dio_mdev_destroy(dio);

	return ret;
}

/**
 * List of supported platform
 */
static const struct platform_device_id fmc_dio_id_table[] = {
	{	/* SPEC compatible */
		.name = "spec-fmc-dio",
		.driver_data = FMC_DIO_BOARD_SPEC,
	},
	{},
};

// FMC DIO platform driver
static struct platform_driver fmc_dio_pl_drv = {
	.probe		= fmc_dio_pl_probe,
	.remove		= fmc_dio_pl_remove,
	.id_table   = fmc_dio_id_table,
	/* No suspend or resume by now */
	.driver		= {
		.name		= "fmc-dio",
		.owner		= THIS_MODULE,
	},
};

/**
 * FMC DIO init function
 *
 * This function is the entry point for the FMC DIO driver.
 *
 * @return 0 if success and a negative error code otherwise
 */
static int fmc_dio_init(void)
{
	int ret;

	ret = fmc_driver_register(&fmc_dio_drv);
	if (ret < 0)
		return ret;
	platform_driver_register(&fmc_dio_pl_drv);
	if (ret < 0)
		fmc_driver_unregister(&fmc_dio_drv);
	return ret;
}

/**
 * FMC DIO init function
 *
 * This function is the exit point for the FMC DIO driver.
 *
 */
static void fmc_dio_exit(void)
{
	platform_driver_unregister(&fmc_dio_pl_drv);
	fmc_driver_unregister(&fmc_dio_drv);
}

module_init(fmc_dio_init);
module_exit(fmc_dio_exit);

/* If no gpio lib is there, this weak applies */

/**
 * FMC DIO GPIO init function
 *
 * @param dev FMC DIO device
 *
 * @return 0 if success and a negative error code otherwise
 */
int __weak fmc_dio_gpio_init(struct fmc_dio *dev)
{
	return 0;
}

/**
 * FMC DIO GPIO exit function
 *
 * @param dev FMC DIO device
 */
void __weak fmc_dio_gpio_exit(struct fmc_dio *dev)
{
	return;
}

MODULE_VERSION(GIT_VERSION);
MODULE_LICENSE("GPL");
